import faiss

def create_empty_faiss_index(dim):
    "Create a new index"
    return faiss.IndexFlatIP(dim)

def get_faiss_index(loc):
    "Get the index"
    try:
        return faiss.read_index(loc + ".index")
    except FileNotFoundError as err:
        raise f"Unable to find {loc}, does the file exist? from {err}"

def add_doc_to_index(index: faiss.IndexFlatIP, embedded_document_text):
    "Add doc to index"
    index.add(embedded_document_text)

def search_index(embedded_query, index: faiss.IndexFlatIP, doc_map, k=5, minimum_score_threshold: float = 0.7, return_scores: bool = True,):
    "Search through the index"
    D, I = index.search(embedded_query, k)
    print(D, I)
    if return_scores:
        value = [{doc_map[idx]: str(score)} for idx, score in zip(I[0], D[0]) if (idx < len(doc_map) and float(score) > minimum_score_threshold)]
    else:
        value = [doc_map[idx] for idx, score in zip(I[0], D[0]) if (idx < len(doc_map) and float(score) > minimum_score_threshold)]
    return value

def save_index(index, loc):
    "Save the index and dataset pickle file to local"
    try:
        faiss.write_index(index, loc + ".index")
    except Exception as err:
        raise err
    
def get_text_queries(loc):
    "Get the text queries"
    try:
        with open(loc + ".txt", 'r') as file:
            return file.read().splitlines()
    except FileNotFoundError as err:
        raise f"Unable to find {loc}, does the file exist? from {err}"
    except Exception as err:
        raise err
    
def save_text_queries(queries, loc):
    "Save the text queries"
    with open(loc + ".txt", 'w') as file:
        file.write('\n'.join(queries))


import os
from typing import Union
import numpy as np
from sentence_transformers import SentenceTransformer

from neuroqrs.index import NeuroIndex
from neuroqrs.genai import get_genai_suggestions

index_caches = {}

class NeuroQRS:
    "NeuroQRS main class"
    def __init__(self, embedding_model = "all-mpnet-base-v2", model_dim: int = 768):
        """Initialize the NeuroQRS
        
        :param embedding_model: SentenceTransformer model to use, defaults to "all-mpnet-base-v2"
        :param model_dim: Dimension of the model, defaults to 768
        :type embedding_model: str
        :type model_dim: int
        """
        self.model = SentenceTransformer(embedding_model)
        self.model_dim = model_dim

    async def quick_query(self, query: Union[str, np.array], namespace: str, k: int = 5):
        """Quickly query the index

        :param query: partial query to search or embedded query
        :param namespace: Namespace associated
        :param k: Number of results to return
        :type query: str
        :type index: str
        :type k: int
        :return: List of results
        :rtype: list
        """

        if namespace not in index_caches:
            index_caches[namespace] = NeuroIndex.from_namespace(namespace, dim=self.model_dim)

        if isinstance(query, str):
            query = self.model.encode([query], convert_to_numpy=True)

        return index_caches[namespace].search(query, k=k)
    

    async def query_and_index(self, query: str, namespace: str, extra_documents_data: dict, fields_info: dict = {}, k: int = 5):
        """Query and index the document (slow, uses even genai to improve context)

        :param query: partial query to search
        :param namespace: Namespace associated
        :param extra_documents_data: extra supporting data, simply dumped to llm prompt
        :param fields_info: fields information (ie what filters are avaliable for the search)
        :param k: Number of results to return
        :type query: str
        :type namespace: str
        :type extra_documents_data: dict
        :type fields_info: dict
        :type k: int
        :return: List of results
        :rtype: list
        """
    
        if namespace not in index_caches:
            index_caches[namespace] = NeuroIndex.from_namespace(namespace, dim=self.model_dim)

        query_embeddings = self.model.encode([query], convert_to_numpy=True)

        # First use genai and prompting to build more results context XD
        # Then index the document
        suggestions = (await get_genai_suggestions(query, extra_documents_data, fields_info)).suggestions
        for suggestion in suggestions:
            index_caches[namespace].add(suggestion, self.model.encode([suggestion], convert_to_numpy=True))

        # Also save the index
        index_caches[namespace].save()

        return index_caches[namespace].search(query_embeddings, k=k)
    
    async def query_maybe_index(self, query: str, namespace: str, extra_documents_data: dict, fields_info: dict = {}, k: int = 5):
        """Query the document quickly for results, if not enough results then index the document and query again (might be slow)
        
        :param query: partial query to search
        :param namespace: Namespace associated
        :param extra_documents_data: extra supporting data, simply dumped to llm prompt
        :param fields_info: fields information (ie what filters are avaliable for the search)
        :param k: Number of results to return
        :type query: str
        :type namespace: str
        :type extra_documents_data: dict
        :type fields_info: dict
        :type k: int
        :return: List of results
        :rtype: list
        """

        # First try to quickly query the index
        results = await self.quick_query(query, namespace, k=k)
        if results and len(results) > 3:
            return results
        else:
            print("Quick query failed, trying to index the document")
            return await self.query_and_index(query, namespace, extra_documents_data, fields_info, k=k)

if __name__ == "__main__":
    neuroqrs = NeuroQRS()
    async def main():
        print(await neuroqrs.query_maybe_index("nike good", "indexeess/nike", extra_documents_data={}, fields_info={"color": ["red", "blue", "green"], "size": ["small", "medium", "large"], "category": ["shoes", "clothes", "accessories", "sandals"], "division": ["men", "women"]}))

    import asyncio
    asyncio.run(main())


import openai, json
from typing import List
from pydantic import BaseModel

client = openai.AsyncOpenAI(
    # defaults to os.environ.get("OPENAI_API_KEY")
    # or set/uncomment below
    #api_key="<api key>"
)

class SuggestedQueries(BaseModel):
    """Suggested Queries model"""
    suggestions: List[str]

async def get_genai_suggestions(partial_query: str, context: dict, fields_info: dict, n_queries: int = 10) -> SuggestedQueries:
    """Get genai suggestions for the given partial query, context and fields_info, Uses latest function call architecture.

    :param partial_query: The partial query for which suggestions are needed
    :param context: The context for the query
    :param fields_info: The fields information
    :param n_queries: Number of queries to suggest, defaults to 10
    :type partial_query: str\
    :type context: dict
    :type fields_info: dict
    :type n_queries: int, optional
    :return: Suggested queries
    :rtype: :class:`SuggestedQueries`
    """
    response = await client.chat.completions.create(
        model="gpt-3.5-turbo",
        messages=[
            {"role": "system", "content": "You are an AI assistant guiding user towards better ecommerce search. You are given a partial query and some context (might not be avaliable), Use it to generate query autocompletion for the user, Important: Only give autocompletions, ie keep the prefix same and suggest change the suffix, Make only coherent meaningful sentences."},
            {"role": "user", "content": f"Call function with {n_queries} autocompletions, Partial Query: {partial_query}, Supporting Data: {json.dumps(context)} and avaliable fields: {json.dumps(fields_info)}"}
        ],
        functions=[
            {
            "name": "get_suggested_queries",
            "description": "Get autocompletion queries for the user based on the partial query and context",
            "parameters": SuggestedQueries.model_json_schema()
            }
        ],
        function_call={"name": "get_suggested_queries"},
        temperature=0.5,
    )

    output = json.loads(response.choices[0].message.function_call.arguments)
    return SuggestedQueries(**output)


import faiss, os
from neuroqrs.utils import create_empty_faiss_index, get_faiss_index, get_text_queries, save_index, save_text_queries, add_doc_to_index, search_index

class NeuroIndex:
    "Faiss powered index for searching through the dataset"
    def __init__(self, namespace: str = "", dim: int = 768):
        """Initialize the NeuroIndex
        
        :param namespace: Namespace for the index, defaults to ""
        :param dim: Dimension of the index, defaults to 768
        :type namespace: str
        :type dim: int
        """
        self.namespace = namespace
        self.dim = dim

        try:
            # Load the index from namespace if it exists
            self.index = get_faiss_index(namespace)
            self.queries = get_text_queries(namespace)
        except Exception as err:
            # Create a new index if it doesn't exist
            self.index = create_empty_faiss_index(dim)
            self.queries = []

    def add(self, document_text: str, embedded_document_text):
        """Add a document to the index and save the query

        :param document_text: Document text
        :param embedded_document_text: Embedded document text
        :type document_text: str
        :type embedded_document_text: list
        """
        add_doc_to_index(self.index, embedded_document_text)
        self.queries.append(document_text)

    def search(self, embedded_query: list, k: int = 5, return_scores: bool = False):
        """Search the index for the given query

        :param embedded_query: Embedded query
        :param k: Number of results to return
        :param return_scores: Return scores or not
        :type embedded_query: list
        :type k: int
        :type return_scores: bool
        :return: List of results
        :rtype: list
        """
        return search_index(embedded_query, self.index, self.queries, k=k, return_scores=return_scores)

    def save(self):
        "Save the index to the namespace"
        # Make folders if they don't exist
        if not os.path.exists(os.path.dirname(self.namespace)):
            if os.path.dirname(self.namespace):
                os.makedirs(os.path.dirname(self.namespace))
    
        save_index(self.index, self.namespace)
        save_text_queries(self.queries, self.namespace)

    @staticmethod
    def from_namespace(namespace: str, dim: int = 768):
        "Create a new index from existing namespace"
        return NeuroIndex(namespace=namespace, dim=dim)